(* Hamming numbers *)
(*
Compile with
ocamlopt nums.cmxa hamming2.ml

The algorithm uses a heap.
First put "1" inside the heap, then iterate:
 - pop the smallest number n from the heap
 - push 2n, 3n, 5n to the heap

We will do 3 versions, each implemented in a class:
 - using ints, it works only for n up to 1237, then overflows.
 - using arbitrary precision numbers "big_int" (OK, not available to C coders)
 - using floats and logs: we record the log10 of the value, and the

*)
open Big_int
(* #load "nums.cma";; *)

class virtual arithmetics = object (self : 'a)
	method virtual unit : 'a
	method virtual add : 'a -> 'a
	method two = self#unit#add self#unit
	method three = self#unit#add self#two
	method five = self#two#add self#three
	method virtual compare : 'a -> int
	method virtual mult : 'a -> 'a
	method virtual string_of : string
	method length = String.length self#string_of
end

class true_big_int = object (self)
	inherit arithmetics
	val v = unit_big_int
	method get = v
	method unit = {< v = unit_big_int >} (* return a new object of this class with v = ... *)
	method compare a = compare_big_int v a#get
	method add a = {< v = add_big_int v a#get >}
	method mult a = {< v = mult_big_int v a#get >}
	method string_of = string_of_big_int v
	method length = String.length self#string_of
end

class int = object (self)
	inherit arithmetics
	val v = 1
	method get = v
	method unit = {< v = 1 >}
	method compare a = compare v a#get
	method add a = {< v = v + a#get >}
	method mult a = {< v = v * a#get >}
	method string_of = string_of_int v
end

class log = object (self)
	inherit arithmetics
	val v = 0.                        (* log scale *)
	val epsilon = 1e-10
	method get = v
	method two  = {< v = log10 2. >} (* log scale *)
	method three = {< v = log10 3. >} (* log scale *)
	method five  = {< v = log10 5. >} (* log scale *)
	method unit = {< v = 0. >}      	(* log scale *)
	method compare a = 
		if abs_float (v -. a#get) < epsilon then 0 else compare v a#get
	method add a = invalid_arg "add should not be called on a log scale"
	method mult a = {< v = v +. a#get >} (* log scale  *)
	method string_of = string_of_float (10. ** v)	(* log scale *)
	method length = truncate v +1
end

(* Change it to test the three versions *)
(* let arith = new true_big_int type hamming = true_big_int *)
(* let arith = new int type hamming = int *)
let arith = new log type hamming = log

let comp_hamming a b = a#compare b

module Heap = Set.Make (struct type t = hamming let compare = comp_hamming end)

let s = ref (Heap.singleton arith#unit)

let push elt =
	s := List.fold_left
		(fun set e -> if not (Heap.mem e set) then Heap.add e set else set)
		!s
		[elt#mult arith#two; elt#mult arith#three; elt#mult arith#five;]

let hamming_numbers = Array.init 200_000
	(fun i ->
		 let a = Heap.min_elt !s in
		 s := Heap.remove a !s;
		 push a;
		 (* Printf.printf "%5i %s\n%!" i (v#string_of); *)
		 (* Printf.printf "%5i %i\n%!" i v#length; *)
		 a)

let scanf format = Scanf.bscanf Scanf.Scanning.stdib format;;
for i=1 to scanf "%i\n" (fun i -> i) do
	scanf "%i\n" (fun i -> Printf.printf "%i\n%!" hamming_numbers.(i-1)#length);
done;;
