(* Hamming numbers *)

(*
 A Hamming number is represented as 4 numbers:
 - its value (redundant with the next three numbers)
 - its valuations in 2, 3 and 5
	 (the valuation in 5 of a integer is the exponent of 5 in its decomposition in prime numbers)

The algorithm uses a heap.
First put "1" inside the heap, then iterate:
 - pop the smallest number n from the heap
 - push 2n, 3n, 5n to the heap

We will do 3 version, each implemented in a class:
 - using ints, it works only for n up to 1237, then overflows.
 - using arbitrary precision numbers "big_int" (OK, not available to C coders)
 - using floats and logs: we record the log10 of the value, and the

*)
open Big_int
#load "nums.cma";;

class virtual arithmetics = object (self : 'a)
	method virtual unit : 'a
	method virtual zero : 'a
	method two = self#unit#add self#unit
	method three = self#unit#add self#two
	method five = self#two#add self#three
	method virtual compare : 'a -> int
	method virtual add : 'a -> 'a
	method virtual mult : 'a -> 'a
	method virtual succ : 'a
	method virtual string_of : string
	method length = String.length self#string_of
end

class true_big_int = object (self)
	inherit arithmetics
	val v = unit_big_int
	method get = v
	method zero = {< v = zero_big_int >}
	method unit = {< v = unit_big_int >}
	method compare a = compare_big_int v a#get
	method add a = {< v = add_big_int v a#get >}
	method mult a = {< v = mult_big_int v a#get >}
	method succ = {< v = succ_big_int v >}
	method string_of = string_of_big_int v
	method length = String.length self#string_of
end

class int = object (self)
	inherit arithmetics
	val v = 1
	method get = v
	method zero = {< v = 0 >}
	method unit = {< v = 1 >}
	method compare a = compare v a#get
	method add a = {< v = v + a#get >}
	method mult a = {< v = v * a#get >}
	method succ = {< v = succ v >}
	method string_of = string_of_int v
end

class log = object (self)
	inherit arithmetics
	val v = 0.                        (* log scale *)
	val epsilon = 1e-10
	method get = v
	method two  = {< v = log10 2. >} (* log scale *)
	method three = {< v = log10 3. >} (* log scale *)
	method five  = {< v = log10 5. >} (* log scale *)
	method zero = {< v = 0. >}
	method unit = {< v = 0. >}      	(* log scale *)
	method compare a = 
		if abs_float (v -. a#get) < epsilon then 0 else compare v a#get
	method add a = {< v = v +. a#get >}
	method mult = self#add            (* log scale  *)
	method succ = {< v = v +. 1. >}
	method string_of = string_of_float (10. ** v)	(* log scale *)
(* 	method string_of = string_of_float v	(\* log scale *\) *)
	method length = truncate v +1
end

let arith = new true_big_int type i = true_big_int
(* let arith = new int type i = int *)
(* let arith = new log type i = log *)

type hamming = i*i*i*i
let comp_hamming (a,_,_,_) (b,_,_,_) = a#compare b

module S = Set.Make (struct type t = hamming let compare = comp_hamming end)

let s = ref (S.singleton (arith#unit,arith#zero,arith#zero,arith#zero))

let push (value,v2,v3,v5) =
	s := List.fold_left
		(fun set e -> if not (S.mem e set) then S.add e set else set)
		!s
		[
			value#mult arith#two  ,v2#succ,v3,v5;
			value#mult arith#three,v2,v3#succ,v5;
			value#mult arith#five ,v2,v3,v5#succ;
		]

let () =
(* for i=1 to 1237 do *)
for i=1 to 200_000 do
	let (v,_,_,_) as a = S.min_elt !s in
	s := S.remove a !s;
	push a;
(* 	Printf.printf "%5i %s\n%!" i (v#string_of); *)
	Printf.printf "%5i %i\n%!" i v#length;
done
