(*
First devise an algorithm for
 - connexity. Here we use a naive O(n^2) method:
	 take one square, look for a neighbor, add it to the list, iterate, check every square has been found.
	 It is doable in O(n) with Tarjan heaps (union-find structure).
 - superposability.
	 Normalize the white configuration by shifting it towards the top left corner as long as possible,
	 then compare it with every normalized symmetry of the black position.
Then it is brute-force search, just careful coding of symmetry.

One could first remark than there is necessary a blank line on one side (row or column)
and a black line on the opposite side, but it's not used here.
*)

(* #load "graphics.cma";; *)
open Graphics;;
open_graph " 1280x1024";;

let height = ref 1 (* bound on the indice i *)
let width  = ref 1 (* bound on the indice j *)

type position = bool array array;;
(* The rectangle to be filled. true for white, false for black. *)

(************** connex? **************)
let is_neighbor (i,j) (k,l) =
	abs(k-i) + abs(l-j) = 1;;

(* return e,rest where test e == true *)
let rec remove test l = match l with
	|[] -> raise Not_found
	|a::q -> if test a then a,q
    else let b,p = remove test q in b,a::p;;

let is_connexe l =
	let rec aux at dt = match at with
			(* at are the elements to be tested, dt the one already tested. dt must not be empty. *)
		| [] -> true
		| _  -> let test a = List.exists (fun d -> is_neighbor a d) dt in
					 try let e,reste = remove test at in aux reste (e::dt) with Not_found -> false in
	match l with
  | [] -> failwith "is_connexe"
  | a::q -> aux q [a];;

(************** superposable? **************)
exception Out

let increment i j =	(* next case *)
	j:= if !j >= pred !width then (incr i; 0) else !j+1;
	if !i >= !height then raise Out;
;;

let decrement i j =
	j:= (if !j<=0 then (decr i; pred !width) else !j-1);;

let copy_position pos = Array.init !height (fun i -> Array.copy pos.(i));;

let is_empty_l pos row =
	let result=ref true and j=ref 0 in
	while !j < !width & !result do
		if pos.(row).(!j) then result := false; incr j;
	done;
	!result;;

let is_empty_c pos column =
	let result=ref true and i=ref 0 in
	while !i < !height & !result do
		if pos.(!i).(column) then result := false; incr i;
	done;
	!result;;

let normalize pos = (* The position pos must not be empty. *)
	let result = copy_position pos in
	while is_empty_l result 0 do
		for i=0 to !height -2 do
			result.(i) <- result.(i+1);
		done;
		result.(pred !height) <- Array.make !width false;
	done;
	while is_empty_c result 0 do
		for i=0 to pred !height do
			for j =0 to !width -2 do
				result.(i).(j) <- result.(i).(j+1);
			done;
			result.(i).(pred !width) <- false;
		done;
	done;
	result;;

let inv i j vect =
	let temp = vect.(i) in vect.(i) <- vect.(j); vect.(j) <- temp;;

let symetrize_Ox pos =
	for i=0 to !height/2 -1 do
		inv i (pred !height -i) pos;
	done;;
(* 	inv 0 3 pos; inv 1 2 pos;; *)

let symetrize_Oy pos =
	for i=0 to pred !height do
		for j=0 to !width/2 -1 do
(* 		inv 0 5 pos.(i); inv 1 4 pos.(i); inv 2 3 pos.(i); *)
			inv j (pred !width -j) pos.(i);
		done;
	done;;

exception Distinct
let superimposed pos1 pos2 =
	let i=ref 0 and j=ref 0 in
(* 	let result = ref (pos1.(!i).(!j) = pos2.(!i).(!j);) in *)
(* 	while !i <= pred !height & !result do *)
	try while true do
		if pos1.(!i).(!j) <> pos2.(!i).(!j) then raise Distinct;
		increment i j;
(* 		result := !result & pos1.(!i).(!j) = pos2.(!i).(!j); *)
	done;
	false	(* We never reach this point in the code *)
	with
	| Distinct -> false
	| Out -> true
(* 	!result;; *)

let superposables pos =
	let whites = normalize pos
	and blacks = Array.map (Array.map not) (copy_position pos) in
  	 (superimposed whites (normalize blacks))
	or (symetrize_Oy blacks; superimposed whites (normalize blacks))
	or (symetrize_Ox blacks; superimposed whites (normalize blacks))
	or (symetrize_Oy blacks; superimposed whites (normalize blacks));;

(************** Graphics **************)
let scale = 15

let draw_rect x y w h =
	let a,b = x+w, y+h in
	moveto x y; lineto a y; lineto a b; lineto x b; lineto x y;;

let draw_position pos (gauche,bas) =
	set_color red; draw_rect !gauche !bas (!width*scale +2) (!height*scale +2);
	for i=0 to pred !height do
		for j=0 to pred !width do
			set_color (if pos.(i).(j) then white else black);
			fill_rect (!gauche + scale*j +1) (!bas + scale*i +1) scale scale;
		done;
	done;
;;

let incr_cursor (x,y) =
	let min_y = (scale * !height +3) in
	if !y >= min_y
	then y:= !y - min_y
	else (y:= size_y () - min_y; x:= !x + (scale * !width +3));
;;


(************** Algorithm core **************)
let next_pos pos =
	let i=ref 0 and j=ref 0 in
	if pos.(0).(0)
	then begin (* first read the else clause *)
		while pos.(!i).(!j) do pos.(!i).(!j) <- false; increment i j; done;
		let a = !i* !width + !j in
		while not pos.(!i).(!j) do increment i j; done;
		pos.(!i).(!j) <- false;
		decrement i j;
		pos.(!i).(!j) <- true;
		for k=1 to a do decrement i j; pos.(!i).(!j) <- true; done; end
	else begin (* simple case : look for a white square and make it move towards pos.(0).(0) *)
		while not pos.(!i).(!j) do increment i j; done;
		pos.(!i).(!j) <- false;
		decrement i j;
		pos.(!i).(!j) <- true; end;;

let solutions = ref [];;

let already_found pos =
	let test = copy_position pos in
	(List.mem test !solutions) or
		(symetrize_Oy test; List.mem test !solutions) or
		(symetrize_Ox test; List.mem test !solutions) or
		(symetrize_Oy test; List.mem test !solutions);;

let main () =
	clear_graph();
	let pos = Array.make_matrix !height !width true in
	let i = ref 0 and j = ref 0 in
	for k=1 to (!height * !width)/2 do pos.(!i).(!j) <- false; increment i j; done;
	let cursor = (ref 0, ref (size_y () - scale * !height -3)) in
	let nb_sols = ref 0 in
(* 	for i=1 to 2704155 do (\*choix de 12 parmi 24, -1 car on ne teste pas la dernire qui plante next_pos*\) *)
	(try while true do
		if superposables pos then begin
				let whites = ref [] and blacks = ref [] in
				for i=0 to pred !height do (*on construit les sommets du graphe dont on teste la connexit*)
					for j=0 to pred !width do
						if pos.(i).(j) then whites:= (i,j) :: !whites else blacks:= (i,j) :: !blacks;
					done;
				done;
				if is_connexe !whites & is_connexe !blacks & not (already_found pos) then ( (*youpi !*)
					draw_position pos cursor;
				  incr_cursor cursor;
					solutions:= (copy_position pos) :: !solutions;
					incr nb_sols;)
		end;
		next_pos pos;
	done; with Out -> ());
	(* The symmetry chacking does not take rotation of 90 into account
		 (possible only for a square), so we have to half the result when we have a square. *)
	Printf.printf "%i\n%!" (if !width = !height then !nb_sols/2 else !nb_sols)
;;

let scanf format = Scanf.bscanf Scanf.Scanning.stdib format;;
for i=1 to scanf "%i\n" (fun i -> i) do
	height := scanf "%i " (fun i -> i);
	width  := scanf "%i\n" (fun i -> i);
	solutions := [];
	main();
done;;
(* Printf.printf "press enter...\n%!";; *)
(* let _ = input_line stdin *)

(* binomial(36,18) = 9075135300 > maxint *)
(* binomial(30,15) = 155117520  < 2^30 *)
