(* Breadth first search with careful coding of the sons of one node. *)

(* ********************************************************************** *)
(* Generic exploration algorithm *)
module Mem_hashtbl = struct
  let create () = Hashtbl.create 1_000
  let add tbl x = Hashtbl.add tbl x ()
  let mem = Hashtbl.mem
end

module Mem = Mem_hashtbl

(** [breadth_first sons get_pos start is_end_pos]
[sons pos] return the list of [moves,pos'] such that taking the move [move] yield the pos [pos'],
[start] is the starting position,
[is_end_pos pos] return [true] iff the algorithm can stop.
@return [None] if the end position can't be reached,
[Some (moves,pos)] if [pos] can be reached with the moves [moves]. *)
let breadth_first
    (sons : 'pos -> ('move * 'pos) list)
    (is_end_pos : 'pos -> bool)
    (start : 'pos) =
  let ht = Mem.create () in
  let to_visit = Queue.create () in
  Queue.add (0,[],start) to_visit;
  let depth = ref 0 in
  let result = ref ([],start) in (* dummy *)
  let walk (moves,pos) =
    if is_end_pos pos then (result:=(List.rev moves, pos); raise Exit);
    ListLabels.iter (sons pos) ~f:(fun (move,f) ->
      if not (Mem.mem ht f)
      then (Mem.add ht f; Queue.add (!depth+1,move::moves,f) to_visit)) in
  try while not (Queue.is_empty to_visit) do walk
		(let p,moves,pos = Queue.pop to_visit in
     if !depth < p then
			 ((* Printf.printf "%i\n%!" p; *)
				 depth := p);
     moves,pos);
		done; None
  with Exit -> Some !result

(* ********************************************************************** *)
(* types and their IO *)
type case =
  | Empty
  | Wall
  | Letter of char
let char_of_case = function
    Empty -> '.'
  | Wall ->  '#'
  | Letter c -> c
let case_of_char = function
  | '.' -> Empty
  | '#' -> Wall
  | 'A'..'Z' as c -> Letter c
  | c -> failwith (Printf.sprintf "caractère non reconnu : %c" c)

type direction = Left | Right | Up | Down

type position = {
  cases:case array array;
  pos_letters:(int*int) array;}
let read_cases input =
  let result = ref [] in
  (try while true do
     let line = input_line input in
(* Printf.printf "read \"%s\"\n%!" line; *)
		 if line = "" then raise End_of_file;
     result := Array.init (String.length line) (fun i -> case_of_char line.[i]) :: !result;
  done with End_of_file -> ());
  Array.of_list (List.rev !result)
let read_pos input =
  let cases = read_cases input in
  let pos_letters = ref [] in
  for i=0 to Array.length cases -1 do
    for j=0 to Array.length cases.(0) -1 do
      match cases.(i).(j) with
        | Letter _ -> pos_letters := (i,j) :: !pos_letters
        | _ -> ()
    done
  done;
  {cases = cases; pos_letters = Array.of_list !pos_letters}

(* ********************************************************************** *)
(* Instanciation of the generic algorithm *)
let copy_matrix a = Array.map Array.copy a
let sons pos =
  let result = ref [] in
  let cases = pos.cases in
  let nb_lines   = Array.length cases in
  let nb_columns = Array.length cases.(0) in
  let incr_l start_l start_c =
    let pos_l = ref start_l in
    while !pos_l < nb_lines   && cases.(!pos_l).(start_c) = Empty do incr pos_l done;
    !pos_l -1 in
  let decr_l start_l start_c =
    let pos_l = ref start_l in
    while !pos_l >= 0         && cases.(!pos_l).(start_c) = Empty do decr pos_l done;
    !pos_l +1 in
  let incr_c start_c start_l =
    let pos_c = ref start_c in
    while !pos_c < nb_columns && cases.(start_l).(!pos_c) = Empty do incr pos_c done;
    !pos_c -1 in
  let decr_c start_c start_l =
    let pos_c = ref start_c in
    while !pos_c >= 0         && cases.(start_l).(!pos_c) = Empty do decr pos_c done;
    !pos_c +1 in
  ArrayLabels.iteri pos.pos_letters
		~f:(fun letter_no (l_letter,c_letter) ->
      let letter_color =
				(match cases.(l_letter).(c_letter)
				 with Letter p -> p | _ -> failwith "error on pos_letters") in
			let add_pos l c direction =
				let new_position =
					{cases = copy_matrix cases;
					 pos_letters = Array.copy pos.pos_letters} in
				new_position.cases.(l_letter).(c_letter) <- Empty;
				new_position.cases.(l).(c) <- Letter letter_color;
				new_position.pos_letters.(letter_no) <- l,c;
				result := ((l_letter,c_letter,direction), new_position) :: !result in
			if l_letter < pred nb_lines   && cases.(l_letter+1).(c_letter  ) = Empty then
				(let l = incr_l (l_letter+1) c_letter in add_pos l c_letter Down);
			if l_letter > 0               && cases.(l_letter-1).(c_letter  ) = Empty then
				(let l = decr_l (l_letter-1) c_letter in add_pos l c_letter Up);
			if c_letter < pred nb_columns && cases.(l_letter  ).(c_letter+1) = Empty then
				(let c = incr_c (c_letter+1) l_letter in add_pos l_letter c Right);
			if c_letter > 0               && cases.(l_letter  ).(c_letter-1) = Empty then
				(let c = decr_c (c_letter-1) l_letter in add_pos l_letter c Left););
  !result

exception Array_for_all
let array_for_all f a =
  try Array.iter (fun elt -> if not (f elt) then raise Array_for_all) a; true
  with Array_for_all -> false

let is_end_pos pos_arrivee pos =
  array_for_all (fun (l,c) -> pos.cases.(l).(c) = pos_arrivee.cases.(l).(c)) pos_arrivee.pos_letters

let scanf format = Scanf.bscanf Scanf.Scanning.stdib format;;

while scanf "%i %i\n" (fun i j -> i<>0 && j<>0) do
  let start_pos = read_pos stdin in
  let end_pos = read_pos stdin in
  match (breadth_first sons (is_end_pos end_pos) start_pos) with
  | None -> Printf.printf "No solution\n%!";
  | Some (moves,pos) -> Printf.printf "%i\n%!" (List.length moves)
done;;
